/*
MLPizzaTimer : a count down timer for a few minutes.

Copyright (C) 2009 by Marcel Lancelle

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, see <http://www.gnu.org/licenses/>.
*/

#include <windows.h>
#include "DiffTime.h"
#include <stdio.h> // for sprintf
#include <math.h>
#include <time.h> // to show end time
#include "../prj/resource.h"

// for ITaskbar Interface
#include <ShObjIdl.h>
UINT g_TaskbarButtonCreated = (UINT)-1;
ITaskbarList3* g_Taskbar = NULL;


#ifndef M_PI
#define M_PI 3.1415926
#endif // M_PI

#define Timer_ID 8000

int winWidth = 230;
int winHeight = 80;

DiffTime diffTime;
int startTimeSeconds = 600;
int currentSeconds, prevSeconds;

bool paused = true;

bool leftMouseButtonDown = false;
int oldMouseY = 0;

void update(HWND hwnd) {
	if (paused) {
		if (startTimeSeconds < 0) startTimeSeconds = 0;
		if (startTimeSeconds > 999*60) startTimeSeconds = 999*60;
	} else {
		float dt = diffTime.getSeconds();
		currentSeconds = startTimeSeconds - (int) (dt + 0.5);

		if (g_Taskbar) g_Taskbar->SetProgressValue(hwnd, startTimeSeconds-currentSeconds, startTimeSeconds);

		if ((currentSeconds < 0) && (prevSeconds >= 0)) { // alarm - time is up
			SetForegroundWindow(hwnd);
			SetActiveWindow(hwnd);
			//Sleep(20);

			// wake up from screensaver by sending a shift keypress (down and up)
			INPUT key[2];
			ZeroMemory(key, sizeof(key));
			for (int i=0; i<2; i++) {
				key[i].type = INPUT_KEYBOARD;
				key[i].ki.wVk = VK_LSHIFT;
			}
			key[1].ki.dwFlags = KEYEVENTF_KEYUP;
			SendInput(2, key, sizeof(INPUT));

			// power up monitor
			DefWindowProc(hwnd, WM_SYSCOMMAND, SC_MONITORPOWER, -1l);
			//SendMessage(hwnd, WM_SYSCOMMAND, SC_MONITORPOWER, -1);
			//SendMessage(HWND_BROADCAST, WM_SYSCOMMAND, SC_MONITORPOWER, -1);

			// sound alarm via pc beeper
			Beep(220, 40); // frequency (Hz), duration (milliseconds)
			Beep(20000, 20);
			Beep(260, 40);

			// sound alarm via default sound
			MessageBeep(-1);

			// flash window on taskbar and caption
			FLASHWINFO fwi;
			fwi.cbSize = sizeof(FLASHWINFO);
			fwi.hwnd = hwnd;
			fwi.dwFlags = FLASHW_TRAY; //FLASHW_ALL;
			fwi.dwTimeout = 500;
			fwi.uCount = 6;
			FlashWindowEx(&fwi);
		}

		prevSeconds = currentSeconds;
	}

	// trigger WM_PAINT
	InvalidateRect (hwnd, NULL, TRUE);
	UpdateWindow(hwnd);
}

void togglePaused(HWND hwnd) {
	paused = !paused;

	if (g_Taskbar) g_Taskbar->SetProgressState(hwnd, paused ? TBPF_PAUSED : TBPF_NORMAL);

	if (paused) {
		startTimeSeconds = currentSeconds;
		KillTimer(hwnd, Timer_ID);
	} else {
		diffTime.start();
		SetTimer(hwnd,Timer_ID, 1000, NULL);
	}

	update(hwnd);
}

// the Window Procedure
LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	HDC hdc;
	PAINTSTRUCT ps;
	HFONT font;

	if (msg == g_TaskbarButtonCreated) {
		::CoCreateInstance(CLSID_TaskbarList, NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&g_Taskbar));
		g_Taskbar->HrInit();
	}

	switch(msg) {
		case WM_PAINT: {
			hdc = BeginPaint (hwnd, &ps); // Gertekontext

			//SetPixel(hdc, 10, 10, RGB(255,0,0)); // Punkt setzen

			int seconds = (paused) ? startTimeSeconds : currentSeconds;
			bool negative = (seconds < 0);
			seconds = abs(seconds);
			int minutes = seconds / 60;
			seconds = seconds % 60;

			if (negative) { // time is up
				SetTextColor(hdc, RGB(paused ? 100 : 255,0,0));

				if ((!paused) && (currentSeconds > -6) && (currentSeconds % 2)) {
					HBRUSH br = CreateSolidBrush(RGB(255,0,0));
					SelectObject(hdc, br);
					Rectangle(hdc, ps.rcPaint.left, ps.rcPaint.top, ps.rcPaint.right, ps.rcPaint.bottom);
					DeleteObject(br);
					SetTextColor(hdc, RGB(255,255,255));
				}
			} else {
				if (paused) SetTextColor(hdc, RGB(100,100,100));
					else SetTextColor(hdc, RGB(0,0,0));
			}

			// line showing percentage
			if (!paused && (currentSeconds > 0)) {
				//int borderWidth = ps.rcPaint.right - ; //GetSystemMetrics(SM_CXBORDER);
				HPEN hLinePen = CreatePen(PS_SOLID, 4, RGB(200, 200, 200));
				SelectObject(hdc, hLinePen);
				MoveToEx(hdc, ps.rcPaint.left, ps.rcPaint.bottom - 3, NULL);
				LineTo(hdc, ps.rcPaint.left + (ps.rcPaint.right-ps.rcPaint.left)*currentSeconds/startTimeSeconds, ps.rcPaint.bottom - 3);
				DeleteObject(hLinePen);
			}

			// analog timers
			if (!paused) {
				COLORREF lightColor = RGB(240, 240, 240);
				COLORREF darkColor = RGB(200, negative ? 0 : 200, negative ? 0 : 200);
				int cx = 23;
				int cy = 23;
				int rad = 20;
				float secDegr = seconds/60.0*2*M_PI;
				HPEN hLinePen = CreatePen(PS_SOLID, 4, darkColor);
				SelectObject(hdc, hLinePen);
				Arc(hdc, cx-rad, cy-rad, cx+rad, cy+rad, cx + rad*sin(secDegr), cy - rad*cos(secDegr), cx, cy-rad);
				DeleteObject(hLinePen);
				hLinePen = CreatePen(PS_SOLID, 4, lightColor);
				SelectObject(hdc, hLinePen);
				Arc(hdc, cx-rad, cy-rad, cx+rad, cy+rad, cx, cy-rad, cx + rad*sin(secDegr), cy - rad*cos(secDegr));
				DeleteObject(hLinePen);

				if (minutes > 0) {
					hLinePen = CreatePen(PS_NULL, 1, RGB(0, 0, 0));
					SelectObject(hdc, hLinePen);
					DeleteObject(hLinePen);
					HBRUSH br = CreateSolidBrush(darkColor);
					SelectObject(hdc, br);
					rad = 17;
					float minDegr = minutes/60.0*2*M_PI;
					Pie(hdc, cx-rad, cy-rad, cx+rad, cy+rad, cx + rad*sin(minDegr), cy - rad*cos(minDegr), cx, cy-rad);
					DeleteObject(br);
				}
			}

			// minutes and seconds
			SetBkMode(hdc, TRANSPARENT);
			font = CreateFont(46, 28, 0, 0,
				FW_NORMAL, FALSE, FALSE, FALSE,
				ANSI_CHARSET, OUT_DEFAULT_PRECIS,
				CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
				DEFAULT_PITCH | FF_SWISS,
				"Verdana");

			SelectObject(hdc, font);

			char buffer[10];
			sprintf(buffer, " %i:%02i", minutes, seconds);
			if (negative) buffer[0] = '-';

			SetTextAlign(hdc, TA_RIGHT);
			TextOut(hdc, 210, -5, buffer, strlen(buffer));
			DeleteObject(font);

// final time
			font = CreateFont(12, 0, 0, 0,
				FW_NORMAL, FALSE, FALSE, FALSE,
				ANSI_CHARSET, OUT_DEFAULT_PRECIS,
				CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
				DEFAULT_PITCH | FF_SWISS,
				"Verdana");

			SelectObject(hdc, font);
			time_t rawtime;
			time(&rawtime); // current time
			rawtime += seconds + 60*minutes;
			struct tm* timeinfo = localtime(&rawtime);
			char* tma = asctime(timeinfo);
			tma += 11; // skip first few infos (weekday, date)
			TextOut(hdc, 207, 34, tma, 8); // 8: only print time (not year)
			DeleteObject(font);

			EndPaint (hwnd, &ps);
			} break;

		case WM_LBUTTONDBLCLK:
			//MessageBox(hwnd, "Double Click", "Mouse", MB_OK);
			togglePaused(hwnd);
			break;

		case WM_LBUTTONDOWN: {
			leftMouseButtonDown = true;
			//POINT mPos; 
			//GetCursorPos(&mPos);
			//oldMouseY = mPos.y;
			//if (oldMouseY > 32000) oldMouseY -= 65536;
			update(hwnd);
			SetCapture(hwnd);
			} break;

		case WM_LBUTTONUP:
			leftMouseButtonDown = false;
			ReleaseCapture();
			break;

		case WM_MOUSEMOVE: {
			//int mouse_x = LOWORD (lParam);
			int mouseY = HIWORD (lParam);
			if (mouseY > 32000) mouseY -= 65536;
			if (paused && leftMouseButtonDown) {
				startTimeSeconds += oldMouseY-mouseY;
				update(hwnd);
			}
			oldMouseY = mouseY;
			} break;

		// needed for wheel
		#if !defined(WM_MOUSEWHEEL)
		# define WM_MOUSEWHEEL 0x020A
		#endif // WM_MOUSEWHEEL
		#if !defined(GET_WHEEL_DELTA_WPARAM)
		# define GET_WHEEL_DELTA_WPARAM(wparam) ((short)HIWORD(wparam))
		#endif // GET_WHEEL_DELTA_WPARAM

		case WM_MOUSEWHEEL:
			if (paused) {
				int delta = GET_WHEEL_DELTA_WPARAM(wParam);

				// speed up when numbers are big
				//int multip = (int) (0.3*log(fabs((float) startTimeSeconds)));
				int multip = (int) (1 + 0.0004*fabs((float) startTimeSeconds));

				if (delta < 0) {
					startTimeSeconds -= multip*60;
				} else if (delta > 0) {
					startTimeSeconds += multip*60;
				}

				update(hwnd);
			}
			break;

		case WM_KEYDOWN :
			switch (wParam) {
				case VK_RETURN:
				case ' ':
					togglePaused(hwnd);
					break;
				case VK_DOWN:
					if (paused) startTimeSeconds -= 10;
					update(hwnd);
					break;
				case VK_UP:
					if (paused) startTimeSeconds += 10;
					update(hwnd);
					break;
				case VK_NEXT: // page down
					if (paused) startTimeSeconds -= 60;
					update(hwnd);
					break;
				case VK_PRIOR: // page up
					if (paused) startTimeSeconds += 60;
					update(hwnd);
					break;
				case VK_F1:
					MessageBox(NULL, "(C) 2009 by Marcel Lancelle", "MLPizzaTimer - About", MB_ICONINFORMATION | MB_OK);
					break;
				case VK_ESCAPE:
					SendMessage(hwnd, WM_CLOSE, 0, 0);
					break;
			}
			break;

		case WM_TIMER:
			update(hwnd);
			break;

		case WM_CLOSE:
			DestroyWindow(hwnd);
			break;
		case WM_DESTROY:
			PostQuitMessage(0);
			break;
		default:
			return DefWindowProc(hwnd, msg, wParam, lParam);
	}
	return 0;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
	// parse CmdLine
	int overrideStartTimeSeconds = atoi(lpCmdLine);
	if (overrideStartTimeSeconds > 0) startTimeSeconds = overrideStartTimeSeconds;

	WNDCLASSEX wc;
	HWND hwnd;
	MSG Msg;

	// Registering the Window Class
	wc.cbSize		= sizeof(WNDCLASSEX);
	wc.style		= CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS; //0;
	wc.lpfnWndProc	= WndProc;
	wc.cbClsExtra	= 0;
	wc.cbWndExtra	= 0;
	wc.hInstance	= hInstance;
	//wc.hIcon		= LoadIcon(NULL, IDI_APPLICATION);
	wc.hIcon		= LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_ICON1));
	wc.hIconSm		= NULL;//wc.hIcon;
	//wc.hIconSm		= (HICON)LoadImage(hInstance, MAKEINTRESOURCE(IDI_ICON1), IMAGE_ICON, 16, 16, 0);
	wc.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszMenuName	= NULL;
	wc.lpszClassName = "mlpizzatimer";

	if (!RegisterClassEx(&wc)) {
		MessageBox(NULL, "Window Registration Failed!", "Error!", MB_ICONEXCLAMATION | MB_OK);
		return 0;
	}

	// Creating the Window
	hwnd = CreateWindowEx(WS_EX_CLIENTEDGE | WS_EX_TOPMOST, wc.lpszClassName, "MLPizzaTimer", WS_BORDER | WS_CAPTION | WS_SYSMENU,
		CW_USEDEFAULT, CW_USEDEFAULT, winWidth, winHeight, NULL, NULL, hInstance, NULL);

	if (hwnd == NULL) {
		MessageBox(NULL, "Window Creation Failed!", "Error!", MB_ICONEXCLAMATION | MB_OK);
		return 0;
	}

	g_TaskbarButtonCreated = ::RegisterWindowMessage("TaskbarButtonCreated");

	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);


	// The Message Loop
	while (GetMessage(&Msg, NULL, 0, 0) > 0) {
		TranslateMessage(&Msg);
		DispatchMessage(&Msg);
	}
	return Msg.wParam;
}
